/* gentest.c -- Copyright (C)  2000-2008 Marco Cesati
 * Sun, 06 Jun 2004 12:42:10 +0200
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#if HAVE_STDLIB_H
#  include <stdlib.h>
#endif

#include <time.h>

#define _GNU_SOURCE
#define __USE_GNU

#if HAVE_STRING_H
#  include <string.h>
#endif

#include <signal.h>

#if HAVE_UNISTD_H
#  include <unistd.h>
#endif

#include "gentest.h"

#define VERSION_AND_COPYRIGHT \
          PACKAGE " version " PACKAGE_VERSION \
          " -- Copyright (C) 2000-2008 Marco Cesati <cesati@uniroma2.it>\n\n" \
          "This is free software; see the source for copying conditions.\n" \
          "There is NO warranty; not even for MERCHANTABILITY or FITNESS " \
          "FOR A PARTICULAR PURPOSE.\n\n"

const char *mktestdefault = "mktest";

static int getint(char const *, int);
static void generate(int, FILE *);

char const extension[] = ".atq";

static void remove_tmp_file(void)
{
    char *p = get_temp_filename();
    if (p && unlink(p)) 
        fprintf(stderr, "WARNING!! Cannot delete the temporary file %s\n"
                "(The temporary file includes all questions and answers!!)\n",
                p);
}

static void graceful_exit(int sig)
{
    fprintf(stderr, "Got signal for \"%s\", exiting\n", strsignal(sig));
    exit(EXIT_FAILURE);
}

static void register_signal_handlers(char *p)
{
    struct sigaction sa;
    int *s, catched[] = { SIGPIPE, SIGSEGV, SIGABRT, SIGINT, SIGQUIT, SIGILL, SIGFPE, 
                      SIGTERM, SIGSYS, SIGXCPU, SIGXFSZ, 0 };
    
    sa.sa_handler = graceful_exit;
    sa.sa_flags = 0;
    for(s=catched; *s; ++s) {
        if (sigaction(*s, &sa, NULL)) {
            fprintf(stderr, "%s: cannot register signal handler for \"%s\"\n", p, strsignal(*s));
            exit(EXIT_FAILURE);
        }
    }
}

int main(int argc, char *argv[])
{
    int l;
    FILE * qf;
    int numform;
    char buffer[255];
    char *filename;
    char *basename;
    char *ev;
    unsigned int seed;

    if (argc < 3) {
        fprintf(stderr, VERSION_AND_COPYRIGHT);
        fprintf(stderr, "Usage: %s <filename>[%s] <number>\n"
                        "\t<filename> specifies the question file\n"
                        "\t<number> specifies the number of forms\n",
                argv[0], extension);
        return EXIT_FAILURE;
    }    

    register_signal_handlers(argv[0]);

    l = atexit(remove_tmp_file);
    if (l) {
        fprintf(stderr, "%s: cannot set exit function, aborted\n", argv[0]);
        exit(EXIT_FAILURE);
    }

    srandom((seed=time(NULL)));

    filename = argv[1];
    l = strlen(filename);
            
    if (filename[l-1] != extension[3] || filename[l-2] != extension[2] 
        || filename[l-3] != extension[1] || filename[l-4] != extension[0]) {
        filename = (char *) malloc(sizeof(char)*(l+5));
        if (!filename) {
                fprintf(stderr, "Memory allocation error\n");
                return EXIT_FAILURE;
        }
        strcpy(filename, argv[1]);
        strcat(filename, extension);
    }

    qf = atq_preprocess(&filename);
    if (!qf) {
        perror(filename);
        return EXIT_FAILURE;
    }

    basename = strdup(get_master_filename());
    l = strlen(basename);
    if (basename[l-1] == extension[3] && basename[l-2] == extension[2] 
        && basename[l-3] == extension[1] && basename[l-4] == extension[0])
        basename[l-4] = '\0';
   
    ev = strrchr(basename, '/');
    if (ev)
        basename = ev+1;
        
    numform = getint(argv[2], strlen(argv[2]));
    if (numform == -1) {
        fprintf(stderr, "Argument `%s' is not a valid integer\n", argv[2]);
        return EXIT_FAILURE;
    }
    
    fprintf(stderr, "Generating %d test sheets from %s ( ", numform, get_master_filename());
    
    readfile(qf);
        
    if (fclose(qf))
        perror(filename);

        
    ev = getenv("MKTESTCMD");
    
    if (!ev || !*ev)
        ev = (char *) mktestdefault;

    sprintf(buffer, "%s %s %s > %s.atd", ev, filename, basename, basename);

    qf = popen(buffer, "w");
    
    if (!qf) {
        perror("mktest");
        return EXIT_FAILURE;
    }

    generate(numform, qf);
    
    l = pclose(qf);

    if (l==-1)
        perror("mktest's pipe");

    return ( l ? EXIT_FAILURE : EXIT_SUCCESS );
}

static int getint(char const *str, int len)
{
    int t = 0;

    if (len > 10 || len < 1)
        return -1;
    
    while (len-- && *str) {
        if (*str > '9' || *str < '0')
            return -1;
        t = t * 10 + (*str++) - '0';
    }
    return t;    
}

static inline int coin(void)
{
    return (random() % 2);
}

static void generate(int numform, FILE *out)
{
#define _MAXINT 999999

    int *ingroup[MAXGROUPS];
    int i;
    int form = 0;
    
    for (i=0; i<ngroup; ++i)
        if (!(ingroup[i] = (int *) calloc(groups_t[i], sizeof(int)))) {
            fprintf(stderr, "Memory allocation error\n");
            exit(EXIT_FAILURE);
        }
    
    
    while (form < numform) {
        int gr;
        int glob = 0;
        int question[MAXGROUPS];
        int nq = 0;
        
        for (gr=0; gr<ngroup; ++gr) {
            int i;
            for (i=0; i<groups_s[gr]; ++i) {
                int j, min=_MAXINT, qs=-1, h;
                for (j=0; j<groups_t[gr]; ++j) {
                    for (h=0; h<i; ++h)
                        if (question[nq-1-h] == j+glob)
                            break;
                    if ((h == i) && (ingroup[gr][j] < min ||
                        (ingroup[gr][j] == min && coin()))) {
                        min = ingroup[gr][j];
                        qs = j;
                    }
                }
                if (qs == -1) {
                    fprintf(stderr, "PANIC: internal error in file %s, "
                           "function %s, line %d\n", __FILE__, __FUNCTION__,
                           __LINE__);
                    exit(EXIT_FAILURE);
                }
                  
                question[nq] = qs+glob;
                ++nq;
                ++ingroup[gr][qs];
            }
            glob += groups_t[gr];
        }
        fprintf(out, "%d: ", form);
        for (i=0; i<nquiz; ++i)
            fprintf(out, "%d ", question[i]);
        fprintf(out, "\n");
        ++form;
    }
    for (i=0; i<ngroup; ++i)
        free(ingroup[i]);
}
