/* preprocess.c -- Copyright (C)  2000-2008 Marco Cesati
 * Thu, 09 Sep 2004 12:36:25 +0200
 */


#include <assert.h>

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#if HAVE_STDLIB_H
#  include <stdlib.h>
#endif

#if HAVE_STRING_H
#  include <string.h>
#endif

#if HAVE_UNISTD_H
#  include <unistd.h>
#endif

#if HAVE_SYS_TYPES_H
#   include <sys/types.h>
#endif

#include "gentest.h"

#define MAX_ATQFILE 100
#define MAX_TMP_PATH 255
#define TMP_FILE_NAME "/tmp/TempATQ-XXXXXX"
#define MAXLINE 4096

static int linepos[MAX_ATQFILE];
static int lineints[2*MAX_ATQFILE];
static char *subfiles[MAX_ATQFILE];
static char *master_file = NULL;
static int num_subfiles = 0;
static char *temp_file = NULL;

char * get_master_filename(void)
{
    return master_file;
}

char * get_temp_filename(void)
{
    return temp_file;
}

void get_file_and_line(int gline, char **name, int *line)
{
    int i, n=(num_subfiles<<1);

    if (!n) {
        *name = master_file;
        *line = gline;
        return;
    }

    for (i=0; i<n; i+=2 )
        if (gline <= lineints[i+1]) 
            break;

    if (i == n || gline < lineints[i]) {
        *name = master_file;
        *line = gline - ( --i < 0 ? 0 : lineints[i] - linepos[i>>1] );
    } else {
        *name = subfiles[i>>1];
        *line = gline - lineints[i] + 1;
    }
}

static int check_groupfile(FILE *f)
{
    int tok, n=0, ll=0;

    restart_atq_scanner(f);
    while ((tok = get_token())) {
        if (tok == ATQFILE) {
           assert(n<=MAX_ATQFILE);
           tok = get_token();
           assert(tok == ARG);
           linepos[n] = _linenum;
           subfiles[n] = strdup(yytext);
#if 0
           fprintf(stderr, "\\atqfile{%s} #%d, line %d\n", subfiles[n], n, linepos[n]);
#endif
           ++n;
        } else
           ll = _linenum;
        if (ll == linepos[n-1]) {
            fprintf(stderr, "Illegal use of \\atqfile command in line %d (it should be alone in the line)\n", ll);
            exit(EXIT_FAILURE);
        }

    }
    return n;
}

static int write_line(int tmp, char *line)
{
    int i=0, len = strlen(line);
    assert(len);
    do {
        int rc = write(tmp, line+i, len-i);
        if (rc < 0)
            return -1;
        i += rc;
    } while (i<len);
    return 0;
}

static int include_atqfile(int tmp, char *atqfn)
{
    int l=0;
    char bline[MAXLINE];
    FILE *f = fopen(atqfn, "r");
   
    if (!f)
        return -1;
    
#if 0
    fprintf(stderr, "Including file %s\n", atqfn);
#endif
    while (!feof(f)) {
       if (!fgets(bline, MAXLINE, f))
           break;
       ++l;
       write_line(tmp, bline);
#if 0
       fprintf(stderr, "[%s,%-3d] %s", atqfn, l, bline);
#endif
    }
    if (fclose(f))
        perror(atqfn);
    return l;
}


static FILE * build_tempfile(FILE *f, char *master, char *name)
{
    char bline[MAXLINE];
    int tfd;
    int l, n, gl;
    FILE *nf;

    strncpy(name, TMP_FILE_NAME, MAX_TMP_PATH);
    tfd = mkstemp(name);
    temp_file = strdup(name);

#if 0
    fprintf(stderr, "tmpname=\"%s\"\n", name);
    fprintf(stderr, "num_subfiles=%d\n", num_subfiles);
#endif
    rewind(f);
    gl = l = 0;
    n = 0;
    while (!feof(f)) {
        if (!fgets(bline, MAXLINE, f))
            break;
        ++l;
        if (l == linepos[n]) {
            int j = n << 1;
            int rc;
            lineints[j]=gl+1;
#if 0
            fprintf(stderr, "[%s,%-3d,%-3d] ===== Starting %s ======\n", master, l, gl, subfiles[n]);
#endif
            rc = include_atqfile(tfd, subfiles[n]);
            if (rc < 0)
                panic("Error in file %s, line %d: cannot access file %s\n", master, l, subfiles[n]);
            gl += rc; 
#if 0
            fprintf(stderr, "[%s,%-3d,%-3d] ===== Ending   %s ======\n", master, l, gl, subfiles[n]);
#endif
            lineints[j+1]=gl;
            ++n;
            continue;
        }
        gl++;
#if 0
        fprintf(stderr, "[%s,%-3d,%-3d] %s", master, l, gl, bline);
#endif
        if (write_line(tfd, bline)) {
            perror(name);
            exit(EXIT_FAILURE);
        }
    } 
#if 0
    fprintf(stderr, "l=%d gl=%d\n", l, gl);
#endif
    if (close(tfd))
        perror(name);
    nf = fopen(name, "r");
    if (!nf) {
        perror(name);
        exit(EXIT_FAILURE);
    }
    return nf;
}

FILE * atq_preprocess(char **pfilename)
{
    char *filename = *pfilename;
    char tmpname[MAX_TMP_PATH];
    int n;

    FILE *f, *nf;
    
    master_file = *pfilename;

    f = fopen(filename, "r");
    if (!f)
	return f;

    num_subfiles = n = check_groupfile(f);

    if (!n)
        return f;

    nf = build_tempfile(f, filename, tmpname);
    if (fclose(f))
        perror(filename);
    *pfilename = strdup(tmpname);
    return nf;
}
