/* pbm.c -- Copyright (C)  2000-2008 Marco Cesati
 * Fri, 18 Feb 2005 13:13:50 +0100
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#if HAVE_STDLIB_H
#  include <stdlib.h>
#endif

#include "rdform.h"


void read_pbm_file(const char *filename, struct image_t *pimage)
{
    FILE *f;
    char buffer[255], *p;
    int nl, lo;
    int white = 0, black = 0;
    int width, height;
    char *image;
            

#define PIXEL(L,O)  image[(L)*width+(O)]

    f = fopen(filename, "r");
    if (!f) {
        perror("");
        exit(EXIT_FAILURE);
    }
    
    fgets(buffer, 255, f);
    if (buffer[0] != 'P' || (buffer[1] != '1' &&
                             buffer[1] != '4')) {
        fprintf(stderr, "Unknown format (not a PBM file)\n");
        exit(EXIT_FAILURE);    
    }    
    
    if (buffer[1] == '1') {
        fprintf(stderr, "ASCII PBM format not yet supported\n");
        exit(EXIT_FAILURE);
    }
    
    if (buffer[2] == '\n') {
        do {
            fgets(buffer, 255, f);
        } while (((buffer[0] == '#') || (buffer[0] == '\n')) && !feof(f));
        p = buffer;
    } else
        p = buffer+2;
        
    if (sscanf(p, "%d %d", &width, &height) != 2) {
        fprintf(stderr, "Cannot get image size\n");
        exit(EXIT_FAILURE);
    }
        

    PRINT("PBM image %dx%d\n", width, height);
    
    image = (char *) malloc(width*height+7);
    
    if (!image) {
        fprintf(stderr, "Memory allocation error\n");
        exit(EXIT_FAILURE);
    }

    nl = 0;
    lo = 0;
    
    p = &PIXEL(0,0);
    
    while (!feof(f)) {
        unsigned char c = (unsigned char) fgetc(f);
        int j;
        for (j=0; j<8; ++j, c<<=1) {
            if (nl==height)
                break;
            if (c & 0x80) {
                ++black;
                *p++ = 1;
            } else {
                ++white;
                *p++ = 0;
            }
            if (++lo == width) {
                ++nl, lo=0;
                break;
            }
        }
    }                

    fclose(f);

    pimage->width = width;
    pimage->height = height;
    pimage->buf = image;

    PRINT("Pixels: black=%d white=%d\n", black, white);
    PRINT("Ink: %2.2f%%\n", (float)black*100.0/(black+white));
#undef PIXEL
} 
