/* statatq.c -- Copyright (C)  2000-2008 Marco Cesati
 * Mon,  5 May 2003 18:38:28 +0200
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#if HAVE_STDLIB_H
#  include <stdlib.h>
#endif
#if HAVE_STRING_H
#  include <string.h>
#endif
#include <assert.h>

#define RDFORMOPTIONS "-q"
#define RDFORMOFFSET 12
    
char const extension[] = ".atd";

const char * rdformdefault = "rdform";

#define MAXQUESTIONS            500
#define MAXQUIZ                  10
#define MAXANSWERSPERQUESTION     5
#define MAXANSWERS      (MAXQUESTIONS*MAXANSWERSPERQUESTION)
#define MAXFORMS               1000
#define BUFFERLEN 8192

char const encode[] = {
    '$', /* -9 */
    '^', /* -8 */
    '/', /* -7 */
    '&', /* -6 */
    '%', /* -5 */
    '#', /* -4 */
    '+', /* -3 */
    '=', /* -2 */
    '-', /* -1 */
    '.', /*  0 */
    '1', /* +1 */        
    '2', /* +2 */        
    '3', /* +3 */        
    '4', /* +4 */        
    '5', /* +5 */        
    '6', /* +6 */        
    '7', /* +7 */        
    '8', /* +8 */        
    '9'  /* +9 */        
};

char scramble[MAXFORMS][MAXQUIZ][MAXANSWERSPERQUESTION];
int stat_answer[MAXQUESTIONS][MAXANSWERSPERQUESTION];
int quiznumber[MAXFORMS][MAXQUESTIONS];

void error_abort(const char * mesg)
{
    fprintf(stderr, "%s\n", mesg);
    exit(EXIT_FAILURE);
}

inline void skip_blanks(char **p, FILE *f)
{
    do {
        while (**p && (**p == ' ' || **p == '\t'))
            ++*p;
        if ((!**p) && (!fgets(*p, BUFFERLEN, f)))
            return;
    } while (**p == ' ' || **p == '\t');
}

void evaluate_results(FILE *f)
{
    char buffer[BUFFERLEN];    
    while (!feof(f)) {
        if (fgets(buffer, BUFFERLEN, f)) {
            char *p = buffer;
            int form = 0;
            int quiz = 0;
                                    
            skip_blanks(&p, f);
            if (!*p || *p == '\n')
                continue;
                
            if (*p < '0' || *p > '9') {
                if (*p++ == '-' && *p++ == '-' && *p == '-')
                    fprintf(stderr, "\nForm number is unreadable\n");
                else
                    fprintf(stderr, 
                        "\nInvalid format: 'digit' expected (1)\n");           
                return;
            }
            
            while (*p && *p >= '0' && *p <= '9') {
                form = form * 10 + (*p) - '0';
                ++p;
            }
                            
            if (!*p || *p++ != ':')
                error_abort("\nInvalid format: ':' expected");
            
            if (!*p || *p++ != ' ') 
                error_abort("\nInvalid format: ' ' expected");
            
            do {
                skip_blanks(&p, f);
                if (*p && *p != '\n') {
                    if (*p++ != '[') 
                        error_abort("\nInvalid format: '[' expected");
                    while (*p && *p >= 'A' && *p <= 'E') {
                        int qn = quiznumber[form][quiz];
                        int an = scramble[form][quiz][*p-'A'];
                        ++stat_answer[qn][an];
                        p++;
                    }                    
                    if (!*p || *p !=']')
                        error_abort("\nInvalid format: ']' expected");
                    ++quiz;
                    p++;
                }
            } while (*p && *p != '\n');
        }  /* if */
    } /* while */
}


void show_stats(FILE * out, int max)
{
    int i, j;
    int tot;
    
    for (i=1; i<=max; ++i) {
        fprintf(out, "Question %d:", i);
        for (j=0, tot=0; j<MAXANSWERSPERQUESTION; ++j)
            tot += stat_answer[i][j];
        fprintf(out, " tot=%d", tot);
        for (j=0; j<MAXANSWERSPERQUESTION; ++j)
            fprintf(out, " %c=%d", 'A'+j, stat_answer[i][j]);
        fprintf(out, "\n");
    }
}

int read_atd_file(FILE *f)
{
    char buffer[BUFFERLEN];
    int tf = 0;
    int maxquestion = -1;
    char *p;
    int i;
    
    if (!fgets(buffer, BUFFERLEN, f)) 
        error_abort("Error reading ATD file (empty)");
    
    if (strncmp(buffer, "ATD ", 4))
        error_abort("Old ATD format doesn't carry enough information");

    p = buffer+4;

    for(i=0; i<19; ++i, ++p)
        if (*p != encode[i])
            error_abort("Unknown encoding format");
    if (!fgets(buffer, BUFFERLEN, f))
        error_abort("Error in ATD file (truncated)");

    for (i=0, p = buffer; *p && *p != '\n'; ++p, ++i) {
        int n = 0, a = 0;
        for (; *p && (*p <= '9' && *p >= '0'); ++p)
            n = n * 10 + *p - '0';
        if (*p != ':' && *p != '\n') 
            error_abort("Error in ATD file (no separator)");
        if (i != n-1)
            error_abort("Error in ATD file (unordered questions)");
        for (p++; *p && (*p != ' ' && *p != '\n'); ++p, ++a) {
            int j;
            for (j=0; j<19; ++j)
                if (encode[j] == *p)
                    break;
            if (j==19)
                error_abort("Error in ATD file (illegal encoding)");
        }
    }            
            
    while (!feof(f)) {
        if (fgets(buffer, BUFFERLEN, f)) {
            char *p = buffer;
            int form = 0;
            int quiz = 0;
                        
            skip_blanks(&p, f);
            if (!*p || *p == '\n')
                continue;
                
            if (*p < '0' || *p > '9')
                error_abort("\nInvalid format in ATD file: "
                        "'digit' expected (1)");
            
            while (*p && *p >= '0' && *p <= '9') {
                form = form * 10 + (*p) - '0';
                ++p;
            }
                            
            if (!*p || *p++ != ':')
                error_abort("\nInvalid format in ATD file: "
                    "':' expected");

            do {
                int tot = 0;
                int ans = 0;
                
                skip_blanks(&p, f);
                if (!*p || *p == '\n')
                    break;
                    
                if (*p < '0' || *p > '9') 
                    error_abort("\nInvalid format in ATD file; "
                        "'digit' expected (2)");
                
                while (*p && *p >= '0' && *p <= '9') {
                    tot = tot * 10 + (*p) - '0';
                    ++p;
                }
                
                if (tot > maxquestion)
                    maxquestion = tot;
                    
                if (!*p || (*p != '(' && *p != '['))
                    error_abort("\nInvalid format in ATD file: "
                        "'(' or '[' expected");
                
                *p++;
                quiznumber[form][quiz] = tot;
                while (*p && (*p != ')' && *p != ']')) {
                    if (*p < 'A' || *p > 'E')
                        error_abort("\nInvalid format in ATD file: "
                               "answer code not in {A,B,C,D,E}");
                    scramble[form][quiz][ans] = *p-'A';
                    ++p;
                    ++ans;
                }
                if (!*p)
                    error_abort("\nInvalid format in ATD file: "
                        "')' or ']' expected");
                ++quiz;
                p++;
            } while (*p);
        } /* if */
        ++tf;
    } /* while */
    return maxquestion;
}

inline void usage(const char * prog)
{
        fprintf(stderr, "Usage: %s <atd file> "
                         "[<pnm file> ...]\n", prog);
        exit(EXIT_FAILURE);
}

int main(int argc, char *argv[])
{
    FILE *atdf;
    char *filename;
    char *basename;
    int l;
    int na = 1;
    int maxquestion = -1;            
    
    memset(scramble, 0, MAXFORMS*MAXQUIZ*MAXANSWERSPERQUESTION*sizeof(char));
    memset(stat_answer, 0, MAXQUESTIONS*MAXANSWERSPERQUESTION*sizeof(int));
    memset(quiznumber, 0, MAXFORMS*MAXQUESTIONS*sizeof(int));
    
    if (argc < 2)
        usage(argv[0]);
        
    if (!argv[na] || !argv[na][0])
        usage(argv[0]);
    
    filename = argv[na];
    l = strlen(filename);
    
    if (filename[l-4] != extension[0] ||
        filename[l-3] != extension[1] ||
        filename[l-2] != extension[2] ||
        filename[l-1] != extension[3]) {
    
        filename = (char *) malloc(sizeof(char)*(strlen(argv[na])+5));
        if (!filename) {
            fprintf(stderr, "Memory allocation error\n");
            return EXIT_FAILURE;
        }
        
        strcpy(filename, argv[na]);
        
        filename[l] = extension[0];
        filename[l+1] = extension[1];
        filename[l+2] = extension[2];
        filename[l+3] = extension[3];
        filename[l+4] = '\0';
        basename = argv[na];
    } else {
        basename = (char *) malloc(sizeof(char)*(strlen(argv[na])+1));
        if (!basename) {
            fprintf(stderr, "Memory allocation error\n");
            return EXIT_FAILURE;
        }
        strcpy(basename, argv[na]);
        basename[strlen(basename)-4] = '\0';
    }
    atdf = fopen(filename, "r");
    if (!atdf) {
        perror(filename);
        return EXIT_FAILURE;
    }    
           
    maxquestion = read_atd_file(atdf);

    assert(maxquestion != -1);
    
    if (fclose(atdf))
        perror(filename);

    if (argc > na+1) {
        char *ev;
        char buffer[BUFFERLEN];
        FILE *pf;
        FILE *of;    

        sprintf(buffer, "%s.stats", basename);
        of = fopen(buffer, "w");
        if (!of) {
            perror(buffer);
            return EXIT_FAILURE;
        }

        ev = getenv("RDFORMCMD");
 
        for (l=na+1; l<argc; ++l) {
            fprintf(stderr, "[%d: %s", l-na, argv[l]);
            if (!ev || !*ev)
                ev = (char *) rdformdefault;
            sprintf(buffer, "%s %s %s", ev, RDFORMOPTIONS, argv[l]);
            pf = popen(buffer, "r");
            evaluate_results(pf);
            pclose(pf); /* FIXME check error code */
            fprintf(stderr, "] ");
        }
        fprintf(stderr, "\n");
        show_stats(of, maxquestion);
        if (fclose(of))
            fprintf(stderr, "Error closing output file\n");
    } else {
        evaluate_results(stdin);
        show_stats(stdout, maxquestion);
    }        
    return EXIT_SUCCESS;
}
